from node import TYPE_OPERATOR, OP_NEG


# Each rule will append its hint message to the following dictionary. The
# function pointer to the apply function of the rule is used as key. The
# corresponding value is a string, which will be used to produce the hint
# message. The string will be processed using string.format().
MESSAGES = {}

# Dictionary used to map a stringified expression tree to the index of the
# best / shortest possibility.
BEST_POSSIBILITY_CACHE = {}

MAX_TREE_DEPTH = 20


class Possibility(object):
    def __init__(self, root, handler, args=()):
        self.root = root
        self.handler = handler
        self.args = args

    def __str__(self):
        if self.handler in MESSAGES:
            msg = MESSAGES[self.handler]

            if callable(msg):
                msg = msg(self.root, self.args)

            return msg.format(self.root, *self.args)

        return repr(self)

    def __repr__(self):
        return '<Possibility root="%s" handler=%s args=%s>' \
                % (self.root, self.handler.func_name, self.args)

    def __eq__(self, other):
        """
        Use node hash comparison when comparing to other Possibility to assert
        that its is the same object as in this one.
        """
        return self.handler == other.handler \
               and hash(self.root) == hash(other.root) \
               and self.args == other.args


def find_possibilities(node, parent_op=None):
    possibilities = []
    handlers = []

    from rules import RULES

    if not node.is_leaf:
        # Traverse through child nodes first using postorder traversal
        for child in node:
            possibilities += find_possibilities(child, node.op)

        # Add operator-specific handlers. Prevent duplicate possibilities in
        # n-ary nodes by only executing the handlers on the outermost node of
        # related nodes with the same operator
        #if node.op in RULES and (node.op != parent_op \
        #                         or node.op not in NARY_OPERATORS):
        #    handlers += RULES[node.op]
        if node.op in RULES:
            handlers += RULES[node.op]

    if node.negated:
        handlers += RULES[OP_NEG]

    # Run handlers
    for handler in handlers:
        possibilities += handler(node)

    return possibilities


def find_parent_node(root, child):
    nodes = [root]

    while nodes:
        node = nodes.pop()

        while node:
            if node.type != TYPE_OPERATOR:
                break

            if child in node:
                return node

            if len(node) > 1:
                nodes.append(node[1])

            node = node[0]


def apply_suggestion(root, suggestion):
    # Apply the suggestion on the sub tree root with the given arguments.
    subtree = suggestion.handler(suggestion.root, suggestion.args)

    # Find the parent node of the sub tree and, if it has a parent node,
    # substitute the new sub tree into the parent's list of child nodes.
    parent_node = find_parent_node(root, suggestion.root)

    if parent_node:
        parent_node.substitute(suggestion.root, subtree)
        return root

    # Apparently the whole tree is updated. Return the new tree instead of the
    # old root node
    return subtree


def pick_best_possibility(parser, possibilities):
    root = parser.root_node

    #print 'pick_best_possibility for:', root

    # Get the final expression
    parser.set_root_node(root.clone())
    result = parser.rewrite_all(index=0)

    def traverse_breadth_first(node, result, depth=0):
        if depth > MAX_TREE_DEPTH:
            print 'warning: Aborting because MAX_TREE_DEPTH is reached.'
            return

        node_expr = str(node)

        if node_expr in BEST_POSSIBILITY_CACHE:
            return BEST_POSSIBILITY_CACHE[node_expr]

        # If the nodes match, there is no rewrite step needed. If the nodes to
        # not converge at all, the step_index will be None as well (since there
        # still is no rewrite step to do).
        step_index = None

        #print '  ' * depth + 'node:', node, 'result:', result, 'equal:', \
        #        node.equals(result)

        if not node.equals(result):
            children = []
            possibilities = find_possibilities(node)

            #print '  ' * depth + ('  ' * depth + '\n').join(map(str,
            #    enumerate(possibilities)))

            for p, possibility in enumerate(possibilities):
                #print '  ' * depth + 'possibility:', possibility, 'on:', node

                # Clone the root node because it will be used in multiple
                # substitutions
                child = apply_suggestion(node.clone(), possibility)

                #print '  ' * depth + 'child:', child

                if child.equals(result):
                    step_index = p
                    break

                children.append(child)

            # If the final expression is not found in the direct children,
            # start searching in the children of the children.
            if step_index is None:
                for c, child in enumerate(children):
                    child_step = traverse_breadth_first(child, result, depth + 1)

                    #print '  ' * depth + 'child_step:', child_step

                    if child_step is not None:
                        step_index = c
                        break

        BEST_POSSIBILITY_CACHE[node_expr] = step_index

        return step_index

    #print '--- start traversal ---'

    step_index = traverse_breadth_first(root, result)

    #print 'step_index:', step_index

    #print '--- cache: ---'
    #print BEST_POSSIBILITY_CACHE

    # Reset the parser's original state
    parser.set_root_node(root)
    parser.find_possibilities()

    return parser.possibilities[step_index]
